importPackage(Packages.com.tivoli.am.fim.trustserver.sts.oauth20);
importPackage(Packages.com.tivoli.am.fim.trustserver.sts.uuser);
importClass(Packages.com.tivoli.am.fim.trustserver.sts.utilities.IDMappingExtUtils);
importClass(Packages.com.tivoli.am.fim.trustserver.sts.utilities.OAuthMappingExtUtils);
importClass(Packages.com.ibm.security.access.user.UserLookupHelper);
importClass(Packages.com.tivoli.am.rba.extensions.PluginUtils);

// constants

//the OAuth client ID used by administrators
var ADMINISTRATOR_CLIENT = "Administrator";
var LOCKING_USER_NAME = "dummyuser";

//client will show registration prompt for u&p
var MSG_BEGIN_REGISTER_FLOW = "Prompt the user to register the card";

//client will display a user-friendly message
var MSG_TOKEN_DISABLED = "This token has been disabled";

//the grant_type the client sends when registering a smart card
var GRANT_TYPE_REGISTER = "register_token";

//after attempting to register a new card: client will show an error message.
//Should it offer to deregister the other card?
var MSG_OTHER_CARD_REGISTERED = "Another card is registered to this user";

//if the user attempts to register a card
var MSG_CARD_ALREADY_REGISTERED = "This card is already registered";

//eg second login of the day
var MSG_GRANT_ACTIVE = "The grant is active and you can log in";

//will want a better system to encode this. Client should try again with
//'&password=<>', not '&user=<>'. DO NOT use a colon because ISAM apparently
//splits on that internally.
function REFRESH_WITH_USERNAME(username) {
	return "Perform a password refresh with user '" + username + "'";
}

var request_type;
var access_token_id = null;
var grant_type = stsuu.getContextAttributes().getAttributeValueByName(
		"grant_type");

var state_id = stsuu.getContextAttributes().getAttributeValueByName("state_id");

var request_type = stsuu.getContextAttributes().getAttributeValueByName(
		"request_type");

var client_id = stsuu.getContextAttributes().getAttributeValueByName(
		"client_id");

var provided_refresh_token = stsuu.getContextAttributes()
		.getAttributeValueByNameAndType("refresh_token",
				"urn:ibm:names:ITFIM:oauth:body:param");

var provided_username = stsuu.getContextAttributes()
		.getAttributeValueByNameAndType("username",
				"urn:ibm:names:ITFIM:oauth:body:param");

var provided_password = stsuu.getContextAttributes()
		.getAttributeValueByNameAndType("password",
				"urn:ibm:names:ITFIM:oauth:body:param");

// the AT and RT will be encoded by the client from the smartcard ID & sent in the request
var smartcard_at = stsuu.getContextAttributes().getAttributeValueByNameAndType(
		"token_at", "urn:ibm:names:ITFIM:oauth:body:param");

var smartcard_rt = stsuu.getContextAttributes().getAttributeValueByNameAndType(
		"token_rt", "urn:ibm:names:ITFIM:oauth:body:param");

// tell the oauth handler what tokens to set in the case of successful authentication.
function setTokensForSuccess(at, rt) {
	if (at === null || at === "") {
		OAuthMappingExtUtils
				.throwSTSUserMessageException("Required information (smart card AT) was not passed in");
	}
	if (rt === null || rt === "") {
		OAuthMappingExtUtils
				.throwSTSUserMessageException("Required information (smart card RT) was not passed in");
	}

	// optionally validate further that the requested AT/RT match a suitable format
	// set the new tokens
	stsuu.addAttribute(new com.tivoli.am.fim.trustserver.sts.uuser.Attribute(
			"urn:ibm:ITFIM:oauth20:custom:token:access_token",
			"urn:ibm:ITFIM:oauth20:custom:token", at));
	stsuu.addAttribute(new com.tivoli.am.fim.trustserver.sts.uuser.Attribute(
			"urn:ibm:ITFIM:oauth20:custom:token:refresh_token",
			"urn:ibm:ITFIM:oauth20:custom:token", rt));
}

// Validate a set of credentials or throw STSUserMessageException.
// The STS does this too for grant_type=password, but only after creating
//a token in the DB. We want to prevent that.
function validateCredentialsOrThrow(username, password) {
	var isAuthenticated = false;
	try {
		isAuthenticated = PluginUtils.isValidUsernamePassword(username,
				password);
	} catch (ex) {
		// Generate an audit event
		PluginUtils.logAuditEvent(username, ex.message, false);
		// Throw an exception in order to stop the flow.
		OAuthMappingExtUtils.throwSTSUserMessageException(ex.message);
	}

	if (isAuthenticated) {
		PluginUtils.logAuditEvent(username, "Authentication successful", true);
	} else {
		PluginUtils.logAuditEvent(username,
				"Invalid username/password. Authentication failed.", false);
		// Throw an exception when authentication failed in order to stop the flow.
		OAuthMappingExtUtils
				.throwSTSUserMessageException("Invalid username/password. Authentication failed.");
	}
}

// how many grants/cards are registered to this user already and active (not locked/disabled)?
function count_active_cards(username) {
	var tokens = OAuthMappingExtUtils.getTokens(client_id, username);
	if (tokens === null) {
		return 0;
	}
	var result = tokens.filter(function(x) {
		return x.isEnabled();
	}).length;
	return result;
}

// debug: dump all stsuu attributes
if (true) {
	IDMappingExtUtils.traceString("Begin tracing attributes");
	var attributes = stsuu.getContextAttributes().getAttributes();
	for (var i = 0; i < attributes.length; i++) {
		var attribute = attributes[i];
		var message = "";
		message += "Attribute: " + attribute.getName() + " has type: "
				+ attribute.getType();
		message += " and values: ";
		for (var j = 0; j < attribute.getValues().length; j++) {
			message += "'" + attribute.getValues()[j] + "' ";
		}
	}
	message += "End tracing attributes";
	IDMappingExtUtils.traceString(message);
}

// User is attempting to register a smart card. If successful, it will be registered.
// The administrator client gets more info & fewer requirements.
if (grant_type == GRANT_TYPE_REGISTER) {
	IDMappingExtUtils.traceString("Grant type register recognised");
	if (provided_username != LOCKING_USER_NAME
			&& count_active_cards(provided_username) > 0) {
		PluginUtils.logAuditEvent(provided_username,
				"Tried to register a card despite having one", false);
		OAuthMappingExtUtils
				.throwSTSUserMessageException(MSG_OTHER_CARD_REGISTERED);
	}

	var at = OAuthMappingExtUtils.getToken(smartcard_at);
	var rt = OAuthMappingExtUtils.getToken(smartcard_rt);

	if (at !== null || rt !== null) {
		// record audit event
		var username;
		if (rt !== null) {
			username = rt.getUsername();
			PluginUtils.logAuditEvent(provided_username,
					"User attempted to register smart card (refresh token "
							+ rt.getId() + ") which was already registered to "
							+ username, false);
		} else {
			username = at.getUsername();
			PluginUtils.logAuditEvent(provided_username,
					"User attempted to register smart card (access token "
							+ at.getId() + ") which was already registered to "
							+ username, false);
		}
		if (client_id == ADMINISTRATOR_CLIENT) {
			// more informative error for admins
			OAuthMappingExtUtils
					.throwSTSUserMessageException("Card already registered to "
							+ username);
		} else {
			OAuthMappingExtUtils
					.throwSTSUserMessageException(MSG_CARD_ALREADY_REGISTERED);
		}
	}

	IDMappingExtUtils
			.traceString("Recognised GRANT_TYPE_REGISTER from client - setting tokens for success && grant_type to password");
	setTokensForSuccess(smartcard_at, smartcard_rt);
	stsuu.getContextAttributes().setAttribute(
			new com.tivoli.am.fim.trustserver.sts.uuser.Attribute("grant_type",
					"urn:ibm:names:ITFIM:oauth:body:param", "password")); // the OAuth engine won't validate the credentials, but does require the grant type to be valid (?)
	if (client_id == ADMINISTRATOR_CLIENT) {
		// make the STS issue a token somehow - if ROPC, it still requires the user's real password
	} else {
		validateCredentialsOrThrow(provided_username, provided_password);
	}

} else if (request_type == "access_token") {
	if (grant_type == "refresh_token") {
		var refresh_token = OAuthMappingExtUtils
				.getToken(provided_refresh_token);

		if (refresh_token === null) {
			// card isn't registered to anyone - begin registration flow. CP will read this as a trigger.
			OAuthMappingExtUtils
					.throwSTSUserMessageException(MSG_BEGIN_REGISTER_FLOW);
		}

		var known_username = refresh_token.getUsername();
		//Check if the user is blockedUser or not. If not, then go ahead.
		//else return error.
		if (known_username !== null) {
			if (known_username.equals(LOCKING_USER_NAME)) {
					OAuthMappingExtUtils
							.throwSTSUserMessageException(MSG_TOKEN_DISABLED);
			}
		}

		var state_id = refresh_token.getStateId();

		var tokens = OAuthMappingExtUtils.getTokens(state_id);

		var access_token = null;
		var token = null; // iterator

		for (var i = 0; i < tokens.length; i++) {
			token = tokens[i];
			if (token.getType().equals("access_token") && !token.isExpired()
					&& token.isEnabled()) {
				access_token = token;
				break;
			}
		}
		if (access_token !== null) {
			IDMappingExtUtils.traceString("access_token is " + access_token
					+ ", enabled is " + access_token.isEnabled());
			OAuthMappingExtUtils.throwSTSUserMessageException(MSG_GRANT_ACTIVE); // login already complete: can use this token for the protected application
		}

		IDMappingExtUtils.traceString("Known token: " + refresh_token);
		IDMappingExtUtils.traceString("Associated username: "
				+ refresh_token.getUsername());
		IDMappingExtUtils.traceString("Is expired: "
				+ refresh_token.isExpired());
		IDMappingExtUtils.traceString("Is enabled: "
				+ refresh_token.isEnabled());

		if (!refresh_token.isEnabled()) {
			OAuthMappingExtUtils
					.throwSTSUserMessageException("This static token has been locked.");
		}

		if (provided_password === null || provided_password === "") {
			// provide the username to the CP so it can prompt with it
			// for password, then refresh with a password
			OAuthMappingExtUtils
					.throwSTSUserMessageException(REFRESH_WITH_USERNAME(known_username));
		}

		// as for refresh_token.isExpired(): the OAuth engine will catch and reject it

		// ensure credentials are correct then continue
		validateCredentialsOrThrow(known_username, provided_password);
	}
	setTokensForSuccess(smartcard_at, provided_refresh_token);
} else {
	OAuthMappingExtUtils.throwSTSException("Unexpected request type '"
			+ request_type + "'");
}
