/*
 * Component Name: isamTokenAuthenticationLibraryClient
 * File Name: Main.c
 *
 * IBM Confidential
 * Object Code Only Source Materials
 * 5747-SM3
 * (c) Copyright International Business Machines Corp. 2018
 * The source code for this program is not published or otherwise divested
 * of its trade secrets, irrespective of what has been deposited with the
 * U.S. Copyright Office.
 */

#include "Main.h"

#if defined _WIN32 || _WIN64
#define strcasecmp strcmp
#endif

/*
 * Process flow:
 * 1. Initialize the ISAM client library.
 * 2. Call authenticate_token with token info for refreshing the token.
 * 3. Finalize ISAM client library.
 */

#if 1

int main(int argc, char *argv[]) {
	bool retVal = false;
	int result = _FAILURE;

	int port = 443;
	bool isSSL = true;

	char *hostname = NULL;
	char *portString = NULL;
	char *isSSLString = NULL;
	char *endpoint = NULL;
	char *clientId = NULL;
	char *authToken = NULL;

	char *errorDescription = NULL;

	if (argc < 9) {
		printf("\nInvalid syntax.");
		printf(
				"\nThe syntax is isamAuth(.exe) -h <hostname> -p <port> -e <endpoint> -s <1 for SSL, 0 for not> -c <client id> -t <Token to login>");
		printf(
				"\n Port is defaulted to 443 and isSSL is defaulted to 1 if not provided");
	} else {
		retVal = areCommandLineParamsValid(argc, argv, &hostname, &portString,
				&endpoint, &isSSLString, &clientId, &authToken);
		if (retVal) {
			if (portString) {
				port = atoi(portString);
			}
			if (isSSLString) {
				isSSL = atoi(isSSLString);
			}

			result = isamInitialize(hostname, port, endpoint, isSSL, clientId,
					getUserCredsCallback, getPasswordCallback);

			if (result == _SUCCESS) {
				result = authenticateToken(NULL, authToken, &errorDescription);
				printf("\nStatus: %s", errorDescription);
				free(errorDescription);
				errorDescription = NULL;
				isamFinalize();
			}
		} else {
			if (hostname)
				free(hostname);
			if (portString)
				free(portString);
			if (endpoint)
				free(endpoint);
			if (isSSLString)
				free(isSSLString);
			if (clientId)
				free(clientId);
			if (authToken)
				free(authToken);
		}
	}

	return 0;
}

bool areCommandLineParamsValid(int argc, char *argv[], char **hostname,
		char **portString, char **endpoint, char **isSSLString, char **clientId,
		char **authToken) {

	bool result = false;
	int i;

	char *key = NULL;

	for (i = 1; i < argc; i = i + 2) {
		asPrintf(&key, argv[i]);

		if (strcasecmp(key, "-h") == 0) {
			asPrintf(hostname, argv[i + 1]);
		} else if (strcasecmp(key, "-p") == 0) {
			asPrintf(portString, argv[i + 1]);
		} else if (strcasecmp(key, "-e") == 0) {
			asPrintf(endpoint, argv[i + 1]);
		} else if (strcasecmp(key, "-s") == 0) {
			asPrintf(isSSLString, argv[i + 1]);
		} else if (strcasecmp(key, "-c") == 0) {
			asPrintf(clientId, argv[i + 1]);
		} else if (strcasecmp(key, "-t") == 0) {
			asPrintf(authToken, argv[i + 1]);
		} else {
			free(key);
			key = NULL;
			break;
		}

		free(key);
		key = NULL;
	}

	/*
	 * Check if all the values are read from the command line.
	 */
	if (i >= argc) {
		/*
		 * Check if all the mandatory parameters are present or not.
		 */
		if (*hostname && endpoint && clientId && authToken) {
			result = true;
		}
	} else {
		printf("\nSome unknown parameter is passed. Exiting.");
	}
	return result;
}

int getPasswordCallback(const char *username, char **password) {
	int retVal = _FAILURE;

	*password = (char *) malloc(512);
	if (*password) {
		printf("\nEnter the password for the user %s: ", username);
		scanf("%s", *password);
		retVal = _SUCCESS;
	}
	return retVal;
}

int getUserCredsCallback(char **username, char **password) {
	int retVal = _FAILURE;

	*username = (char *) malloc(512);
	*password = (char *) malloc(512);

	if (*username && *password) {
		printf("\nEnter your username: ");
		scanf("%s", *username);
		printf("\nEnter your password: ");
		scanf("%s", *password);
		retVal = _SUCCESS;
	}
	return retVal;
}

char *removeSingleQuotesFromString (const char *string) {
	int length = strlen (string);
	char *newString = (char*) malloc (length + 1);
	if (newString) {
		strncpy (newString, (string + 1), (length - 2));
	}

	return newString;
}

void asPrintf(char **dest, const char *source) {
	char *tempString = NULL;
	tempString = (char *) malloc(strlen(source) + 1);
	strcpy(tempString, source);
	printf("\nasPrintf: tempString is %s", tempString);

	/*
	 * Remove the single quotes if present. For some reason the code is
	 * seeing single quotes in the DEBUG run but not in RELEASE run. My
	 * release build run on Windows is failing too. I think it may be for
	 * the same reason.
	 */
	if (strstr(tempString, "\'")) {
		*dest = removeSingleQuotesFromString(tempString);
		printf("\nasPrintf: dest received is %s", *dest);
		free(tempString);
		tempString = NULL;
	} else {
		*dest = tempString;
	}
}


#else
/*
 * Process flow:
 * 1. Initialize the ISAM client library.
 * 2. Call authenticate_token with token info for refreshing the token.
 * 	2a. If call fails, check the error code and take appropriate action.
 * 		2a1. If ISAM asks to refresh password, ask the user to provide the
 * 		password and call authenticate_token with password.
 *
 * 		2a2. If ISAM asks the user to register, get the username and password
 * 		for the user and call authenticate_token with both username and
 * 		password.
 *
 * 		2a3. If any other error code comes back, display the error reason and
 * 		do not allow the user to login.
 * 	2b. If call pass, take next step/s.
 * 3. Finalize ISAM client library.
 */
int main() {
	int result = _SUCCESS;
	char *errorDescription = NULL;
	char *token = NULL;
	char *username = NULL;
	char *password = NULL;

	int actionCode = _REQ_REFRESH_AUTH;
	/*
	 * Initialize the library so that ISAM contact information and/or
	 * scopes etc. are already set.
	 */
	result = isamInitialize(hostname, port, endpoint, true, clientId);
	if (result == _SUCCESS) {
		token = getToken();

		result = authenticateToken(NULL, NULL, scope, token, actionCode,
				&errorDescription);

		/*
		 * Repeat the loop, until the user is successfully logged in OR
		 * user has pressed cancel -- TODO:: Cancel logic.
		 */
		while ((result != _SUCCESS) && (result != _CANNOT_CONTINUE)) {

			if (errorDescription) {
				/*
				 * Try getting username out of the errorDescription
				 */
				username = getUserFromErrorString(errorDescription);
				free(errorDescription);
				errorDescription = NULL;
			}

			switch (result) {
				case _REGISTER_USER:
				printf("\nThe card is not registered. Please register it.");

				username = getUserName();
				password = getPassword(username);
				actionCode = _REQ_REGISTER_USER;

				result = authenticateToken(username, password, scope, token,
						actionCode, &errorDescription);

				printf("\nERROR: %s\n", errorDescription);
#ifdef DEBUG
				printf(
						"\nERROR:: While registering the card: error no. is %d\n",
						result);
#endif
				break;

				case _REFRESH_PASSWORD:
				/*
				 * Fetch the username from the user creds struct.
				 */
				printf("\nUser login expired. Please login again.");
				password = getPassword(username);
				actionCode = _REQ_REFRESH_PASSWORD;

				result = authenticateToken(NULL, password, scope, token,
						actionCode, &errorDescription);

				printf("\nERROR: %s\n", errorDescription);
#ifdef DEBUG
				printf(
						"\nERROR:: While registering the card: error no. is %d\n",
						result);
#endif
				break;

				case _USER_REGISTERED:
				/*
				 * It should not come to this flow. If it does, something
				 * somewhere has gone wrong in the library.
				 */
				printf(
						"\nERROR:: This user already has a card registered. Cannot register another card.\n");
				result = _CANNOT_CONTINUE;
				break;

				case _CARD_REGISTERED:
				printf(
						"\nERROR:: Card already in use. Cannot register again.\n");
				result = _CANNOT_CONTINUE;
				break;

				case _ACTIVE_GRANT:
				printf("\nUser is already logged in.\n");
				result = _SUCCESS;
				break;
				case _CARD_DISABLED:
				printf("\nCard disabled. Contact your administrator.\n");
				result = _CANNOT_CONTINUE;
				break;
				case _TOKEN_LOCKED:
				printf("\nCard is locked. Cannot login.\n");
				result = _CANNOT_CONTINUE;
				break;
				default:
				printf("\nERROR:: Unknown error.\n");
				result = _CANNOT_CONTINUE;
				break;
			}

			if (errorDescription) {
				free(errorDescription);
				errorDescription = NULL;
			}
		}
		if (result == _SUCCESS) {
			printf("\nCall is successful.\n");
		}

	} else {
		printf("\nERRROR:: Unable to initialize the client library.\n");
	}
	/*
	 * Finalize the library and exit.
	 */
	isamFinalize();

	/*
	 * Free errorDescription
	 */
	if (errorDescription) {
		free(errorDescription);
		errorDescription = NULL;
	}

	if (token) {
		free(token);
		token = NULL;
	}

	if (username) {
		free(username);
		username = NULL;
	}

	if (password) {
		free(password);
		password = NULL;
	}
	return 0;
}

char* getUserName() {
	char *username = NULL;
	username = (char*) malloc(MAX_PARAM_SIZE);
	if (username) {
		printf("\nEnter user name: ");
		scanf("%s", username);
	}

	return username;
}

char* getPassword(char *username) {
	char *password = NULL;
	password = (char*) malloc(MAX_PARAM_SIZE);
	if (password) {
		printf("\nEnter password for user %s: ", username);
		scanf("%s", password);
	}

	return password;
}

char* getToken() {
	char *token = NULL;
	token = (char*) malloc(MAX_PARAM_SIZE);
	if (token) {
		printf("\nEnter token: ");
		scanf("%s", token);
	}

	return token;
}

char* removeQuotes(const char *username) {
	char *tempString = NULL;
	char *newString = NULL;

	if (username) {
		tempString = strdup(username);

		newString = (char*) malloc(strlen(tempString) - 1);
		if (newString) {
			strncpy(newString, (tempString + 1), (strlen(tempString) - 2));
#ifdef DEBUG
			printf("\nUsername is %s", newString);
#endif
		}

		free(tempString);
		tempString = NULL;
	}
	return newString;
}

char* getUserFromErrorString(const char *errorDescription) {
	char *newUsername = NULL;
	char *username = NULL;

	if (errorDescription) {
		newUsername = strchr(errorDescription, '\'');
		if (newUsername) {
			username = removeQuotes(newUsername);
		}
	}
	return username;
}

#endif

