/*
 * Component Name: isamTokenAuthenticationLibrary
 * File Name: isamTokenAuthInit.c
 *
 * IBM Confidential
 * Object Code Only Source Materials
 * 5747-SM3
 * (c) Copyright International Business Machines Corp. 2018
 * The source code for this program is not published or otherwise divested
 * of its trade secrets, irrespective of what has been deposited with the
 * U.S. Copyright Office.
 */
#include "isamTokenAuthApi.h"

/*
 * This method is used to insert http:// OR https:// in the hostname and
 * then append it with the full path to get the tokens from ISAM.
 */
char* getEndpointUrl(const char *hostname, const int port,
		const char *tokenEndpoint, const int isSSL) {
	int hostnameLen = 0;

	char *hostnameBuffer = NULL;
	char *tempHostname = NULL;

	char *endpointUrl = NULL;

	if (hostname && tokenEndpoint) {
		if (_FAILURE == check_url(hostname)) {
			/*
			 * Setting the length to the higest possible length, which will
			 * be for https (because of an extra 's'). This buffer will
			 * suffice for both http:// OR https://
			 */
			hostnameLen = strlen("https://") + strlen(hostname) + 1;
			hostnameBuffer = (char *) malloc(hostnameLen * sizeof(char));
			if (hostnameBuffer) {
				tempHostname = hostnameBuffer;

				if (isSSL == true) {
					strcpy(hostnameBuffer, "https://");
				} else {
					strcpy(hostnameBuffer, "http://");
				}
				strcat(hostnameBuffer, hostname);
			}
		} else {
			/*
			 * http:// OR https:// is present in the hostname. No need to
			 * make any modifications to the hostname.
			 */
			as_printf(&hostnameBuffer, hostname);
		}

		endpointUrl = create_full_endpoint_url(hostnameBuffer, port,
				tokenEndpoint);

		if (tempHostname) {
			free(tempHostname);
			tempHostname = NULL;
		}
		if (hostnameBuffer) {
			free(hostnameBuffer);
			hostnameBuffer = NULL;
		}
	} else {
#ifdef DEBUG
		printf("\nInvalid parameters passed.");
#endif
	}
	return endpointUrl;
}
